package MIB::CGI::Template;

use strict;
use warnings;
use utf8;
use Template::Multilingual;
use I18N::AcceptLanguage;
use CGI;
use Encode;
use Cache::FileCache;
use File::Path::Set;
use Config::Any::Merge;
use Config::Any::INI;
use Config::Tiny; # Config::Any::INI braucht das
use File::Basename;

sub new {
	my($class, $config) = @_;
	my %self;

	# Verzeichnis für nach Perl übersetzte Templates (Template-Toolkit COMPILE_DIR)
	$self{'compile'} = $config->{'compile_dir'} if $config->{'compile_dir'};

	# Oberstes Template (,,Wrapper''), das verarbeitet wird
	$self{'wrapper'} = $config->{'wrapper'} // 'wrapper';

	# Include-Path. Das Benutzerverzeichnis, in dem die Datei liegt, wird jeweils hinzugefügt.
	# Man braucht das nur für Dateien, die der Wrapper nachzieht (und den Wrapper selber)
	$self{'include'} = $config->{'include_dirs'} if $config->{'include_dirs'} ;

	# Funktion, um aus dem HTTP-Accept-Language-Parameter und den vorhandenen Sprachen die ausgelieferte Sprache zu bestimmen.
	# strict = 0 besagt, daß als Ersatz für de-at ggf. auch de benutzt wird.
	$self{'acceptor'} = I18N::AcceptLanguage->new(strict => 0);

	# Cache für Template-Ausgaben und deren TTL.
	if ($config->{'cache_dir'}) {
		$self{'cache'} = Cache::FileCache->new({ cache_root => $config->{'cache_dir'} });
		$self{'ttl'} = $config->{'ttl'} // 20;
	}

	bless \%self, $class;
}

sub process {
	my ($self, $cgi) = @_;
	my $acceptor = $self->{'acceptor'};
	my $compile  = $self->{'compile'};
	my $cache    = $self->{'cache'};
	my @include  = @{$self->{'include'} // []};
	my $file     = $ENV{SCRIPT_FILENAME};
	my $ttl      = $self->{'ttl'};
	my $url      = $ENV{REQUEST_URI};
	my $docroot  = $ENV{DOCUMENT_ROOT};
	my $output;
	my $wrapper  = $self->{'wrapper'};

	# Auch der wrapper, der direkt an Template::Multilingual::processs() übergeben wird, darf kein absoluter Filename sein
	my ($wrapper_dirname, $wrapper_basename) = (dirname($wrapper), basename($wrapper));
	push @include, $wrapper_dirname;

	# Konfigurations-Cookie auslesen
	my %cookie_config = $cgi->cookie('config');

	# Mittels File::Path::Set/Config::Any::Merge die Benutzer-Konfiguration aus dem www-Baum auslesen
	my @config_files = File::Path::Set->new( base => $docroot, file => 'config.inc')->file_set($file);
	# Falls existent, Konfigurationsdatei im Wrapper-Verzeichnis davorstellen
	my $wrapper_config = $wrapper_dirname . "/" . "config.inc";
	unshift @config_files, $wrapper_config if -r $wrapper_config;
	my $user_config  = Config::Any::Merge->load_files( { files => \@config_files, force_plugins => [ 'Config::Any::INI' ] } );

	# Defaultsprache ist de, außer, in der Benutzer-Konfigurations steht ein default_lang
	my $default_lang = $user_config->{'default_lang'} if $user_config->{'default_lang'} and $user_config->{'default_lang'} =~ m/^[A-Za-z]+$/;
	$default_lang ||= 'de';

	# Vom HTTP-Client angeforderte Sprache(n) (daraus und aus der Liste der im multilingualen Template vorkommenden Srachen wird bestimmt, was ausgeliefert wird)
	# Prioritätsliste: CGI-Parameter, Cookie, HTTP-Header, Default
	# (der Wert aus dem CGI-Parameter wird ggf. später in den Cookie zurückgespeichert)
	my $lang = $cgi->param('language') || $cookie_config{'language'} || $ENV{HTTP_ACCEPT_LANGUAGE} || $default_lang;

	$user_config->{show_news} //= 1;
	$user_config->{lang_selector} //= 1;

	# Der Benutzer kann Zeichensatz und Content-Type angeben.
	# charset besagt, was im HTTP-Header angegeben wird und in welchen Zeichensatz die Template-Ausgabe vor der Auslieferung umgewandelt wird.
	# type gibt den Content-Type an, der im HTTP-Header angegeben wird.
	# encoding besagt, welches Encoding Template-Toolkit in Dateien ohne BOM annimmt.
	my $charset = $user_config->{'charset'} if $user_config->{'charset'} and $user_config->{'charset'} =~ m/^[0-9A-Z-]+$/;
	my $type = $user_config->{'type'} if $user_config->{'type'} and $user_config->{'type'} =~ m=^[a-z]+/[a-z-]+$=;
	my $encoding = $user_config->{'encoding'} if $user_config->{'encoding'} and $user_config->{'encoding'} =~ m/^[0-9A-Z-]+$/;

	# Vom Dateinamen vorne Document Root abschneiden (das kommt in den Include-Path).
	# Dadurch können die Filenamen in PROCESS/INCLUDE dann relativ zum Benutzerverzeichnis sein und man braucht keine mit / oder ./ beginnenden Filenamen (was Template-Toolkit ABSOLUTE bzw. RELATIVE nennt) zuzulassen
	die "SCRIPT_FILENAME $file doesn't start with DOCUMENT_ROOT $docroot\n" unless substr($file, 0, length($docroot) + 1) eq $docroot."/";
	$file = substr($file, length($docroot) + 1);
	push @include, $docroot;

	# Durch eine Template-Multilingual-Erweiterung kann language (LANGUAGE_VAR) eine Funktion sein und bekommt dann als Argument eine Liste der tatsächlich vorkommenden Sprachen.
	# Dadurch, wie sub in Perl funktioniert, ist es OK, in de Funktion $acceptor und $lang/$default_lang zu verwenden; das bezeichnet die jeweils zum Zeitpunkt der Funktionsdefinition sichtbaren entsprechenden Variablen.
	my $vars = {
			file     => $file,
			cgi      => $cgi,
			encoding => $encoding,
			config   => $user_config,
			language => sub { return $acceptor->accepts($lang, [@_]) || $default_lang; },
	};

	# Gegebenenfalls den Teplate-Output aus dem Cache nehmen
	my ($cached, $key);
	if ($cache) {
		$cached = ($cgi->param('cache') || 'yes') ne 'no' ? 1 : 0;
		$key = $cgi->url(-path_info=>1, -query=>1) . '&language=' . $lang ;
		$output = $cache->get($key) if $cached; 
	}

	if (not defined $output) {
		# Template::Multilingual-Objekt mit passendem Include-Path erzeugen.
		# Der Trick, den Provider manuell zu erzeugen, beim Erzeugen des Template-Objektes anzugeben und darüber den Pfad nachträglich zu ändern, funktioniert nicht, weil beim Anlegen des Providers dann der falsche (nicht-Multilingual-)Parser erzeugt wird.
		# LANGUAGE_LIST_VAR ist eine Erweiterung; dieser Variablen wird die Liste der vorkommenden Sprachen zugewiesen; dadurch kann man im Template dann eine HTML-Sprachauswahl relaisieren, die nur die vorkommenden Sprachen auflistet.
		my %template_config;
		$template_config{'INCLUDE_PATH'} = \@include;
		$template_config{'ENCODING'} = $encoding if $encoding;
		$template_config{'COMPILE_DIR'} = $compile if $compile;
		$template_config{'LANGUAGE_VAR'} = 'language';
		$template_config{'LANGUAGE_LIST_VAR'} = 'languages';
		my $template = Template::Multilingual->new(\%template_config);

		# In der Konfigurationsdatei im www-Baum kann (aus einer von uns vorgegebenen Auswahl) ein Wrapper ausgewählt werden.
		my $style = $user_config->{'style'} if $user_config->{'style'} and $user_config->{'style'} =~ m/^\w+$/;
		if ($style) { # testen, ob es im Wrapper-Verzeichnis eine passende Datei gibt
			$wrapper_basename .= '.' . $style if -r $wrapper . '.' . $style;
		}

		# Die als URL-Parameter angegebene Sprache ggf. im Cookie speichern
		$cookie_config{'language'} = $cgi->param('language') if $cgi->param('language');
		my $cookie = $cgi->cookie(-name => 'config', -value => \%cookie_config);

		my ($header, $content);
		$header = $cgi->header(-cookie => $cookie, -charset => $charset, -type => $type);
		$template->process($wrapper_basename, $vars, \$content) or die $template->error() . "\n";
		$output = $header . $content;

		# Ergebnis ggf. in den Cache zurückschreiben
		if ($cache) {
			$cache->set($key, $output, $ttl) if $cached;
		}
	}

	# Ausliefern
	if ($charset) {
		print encode($charset, $output);
	} else {
		print $output;
	}
}

1;
